"""
Module providing access to raw CPU data.
"""

from libcpuid.errors import CLibraryError
from libcpuid._libcpuid_cffi import (  # pylint: disable=no-name-in-module, import-error
    lib,
    ffi,
)


class CPURawData:
    """
    Class holding raw data about a single logical CPU.
    """

    def __init__(self, c_cpu_raw_data):
        self._c_cpu_raw_data = c_cpu_raw_data

    @property
    def c_cpu_raw_data(self):
        """Returns the underlying C structure."""
        return self._c_cpu_raw_data

    def serialize(self, filename: str):
        """Exports the raw data into a provided file."""
        lib.cpuid_serialize_raw_data(self._c_cpu_raw_data, filename.encode())

    @classmethod
    def from_current_cpu(cls):
        """Creates a :class:`CPURawData` instance by running `cpuid` on the current CPU."""
        c_cpu_raw_data = ffi.new("struct cpu_raw_data_t *")
        if lib.cpuid_get_raw_data(c_cpu_raw_data) != 0:
            raise CLibraryError
        return cls(c_cpu_raw_data)

    @classmethod
    def from_cpu_core(cls, logical_cpu: int):
        """Creates a :class:`CPURawData` instance by running `cpuid` on the specified CPU."""
        c_cpu_raw_data = ffi.new("struct cpu_raw_data_t *")
        if lib.cpuid_get_raw_data_core(c_cpu_raw_data, logical_cpu) != 0:
            raise CLibraryError
        return cls(c_cpu_raw_data)

    @classmethod
    def from_file(cls, filename: str):
        """Creates a :class:`CPURawData` instance by parsing the data from a provided file."""
        c_cpu_raw_data = ffi.new("struct cpu_raw_data_t *")
        if lib.cpuid_deserialize_raw_data(c_cpu_raw_data, filename.encode()) != 0:
            raise CLibraryError
        return cls(c_cpu_raw_data)


class CPURawDataArray:
    """
    Class holding raw data about multiple CPUs.
    Instances of this class can be indexed like lists. each item
    (of type :class:`CPURawData`) holds data about one **logical** CPU.
    """

    def __init__(self, c_cpu_raw_data_array):
        self._c_cpu_raw_data_array = c_cpu_raw_data_array
        self._cpu_raw_data_list = [
            CPURawData(c_cpu_raw_data)
            for c_cpu_raw_data in self._c_cpu_raw_data_array.raw[
                0 : self._c_cpu_raw_data_array.num_raw
            ]
        ]

    def __del__(self):
        lib.cpuid_free_raw_data_array(self._c_cpu_raw_data_array)

    def __getitem__(self, index: int) -> CPURawData:
        return self._cpu_raw_data_list[index]

    def __len__(self) -> int:
        return len(self._cpu_raw_data_list)

    @property
    def c_cpu_raw_data_array(self):
        """Returns the underlying C structure."""
        return self._c_cpu_raw_data_array

    def serialize(self, filename: str):
        """Exports the raw data array into a provided file."""
        lib.cpuid_serialize_all_raw_data(self._c_cpu_raw_data_array, filename.encode())

    @classmethod
    def from_all_cpus(cls):
        """Creates a :class:`CPURawDataArray` instance by running `cpuid` on all CPUs."""
        c_cpu_raw_data_array = ffi.new("struct cpu_raw_data_array_t *")
        if lib.cpuid_get_all_raw_data(c_cpu_raw_data_array) != 0:
            raise CLibraryError
        return cls(c_cpu_raw_data_array)

    @classmethod
    def from_file(cls, filename: str):
        """Creates a :class:`CPURawDataArray` instance by parsing the data from a provided file."""
        c_cpu_raw_data_array = ffi.new("struct cpu_raw_data_array_t *")
        if (
            lib.cpuid_deserialize_all_raw_data(c_cpu_raw_data_array, filename.encode())
            != 0
        ):
            raise CLibraryError
        return cls(c_cpu_raw_data_array)



"""
The libcpuid package defines Python bindings to the
libcpuid C library, which provides CPU identification.
"""

from libcpuid._libcpuid_cffi import (  # pylint: disable=no-name-in-module, import-error
    ffi,
    lib,
)
from libcpuid import enums
from libcpuid.errors import CLibraryError
from libcpuid._utils import c_string_to_str


def version() -> str:
    """Returns the version of the libcpuid library."""
    return c_string_to_str(lib.cpuid_lib_version())


def cpuid_present() -> bool:
    """Checks if the `cpuid` instruction is supported."""
    return bool(lib.cpuid_present())


def get_total_cpus() -> int:
    """Returns the total number of logical CPU threads."""
    return lib.cpuid_get_total_cpus()


def get_vendor() -> enums.CPUVendor:
    """Returns the CPU vendor of the current CPU."""
    try:
        return enums.CPUVendor(lib.cpuid_get_vendor())
    except ValueError:
        return enums.CPUVendor.UNKNOWN


def get_cpu_list(vendor: enums.CPUVendor) -> list[str]:
    """Gets a list of CPU :meth:`codenames <info.CPUInfo.cpu_codename>` for a specific vendor."""
    c_cpu_list = ffi.new("struct cpu_list_t *")
    lib.cpuid_get_cpu_list(vendor, c_cpu_list)
    if c_cpu_list.num_entries == 0:
        raise CLibraryError
    cpu_list = [
        c_string_to_str(name) for name in c_cpu_list.names[0 : c_cpu_list.num_entries]
    ]
    lib.cpuid_free_cpu_list(c_cpu_list)
    return cpu_list


def exec_cpuid(
    eax: bytes,
    ebx: bytes = 4 * b"\x00",
    ecx: bytes = 4 * b"\x00",
    edx: bytes = 4 * b"\x00",
) -> tuple[bytes, bytes, bytes, bytes]:
    """Executes the `cpuid` instruction with the given content of input registers."""
    regs = ffi.new(
        "uint32_t[4]",
        [
            int.from_bytes(eax),
            int.from_bytes(ebx),
            int.from_bytes(ecx),
            int.from_bytes(edx),
        ],
    )
    lib.cpu_exec_cpuid_ext(regs)
    return tuple(reg.to_bytes(4) for reg in regs)



"""
The main module, providing a class that holds CPU information.
"""

from typing import Optional
from libcpuid import enums
from libcpuid.sgx import SGX
from libcpuid.raw import CPURawData, CPURawDataArray
from libcpuid.errors import CLibraryError
from libcpuid._utils import c_string_to_str, optional_int
from libcpuid._libcpuid_cffi import (  # pylint: disable=no-name-in-module, import-error
    lib,
    ffi,
)


class CPUInfo:  # pylint: disable=too-many-public-methods
    """
    The main class, holding information about a CPU
    in human-friendly format.
    """

    def __init__(self, c_cpu_id):
        self._c_cpu_id = c_cpu_id
        self._features = None
        self._detection_hints = None

    @classmethod
    def from_c(cls, c_cpu_id):
        """Creates a :class:`CPUInfo` instance from the corresponding C structure."""
        cpu_info = cls(c_cpu_id)
        if cpu_info.architecture == enums.CPUArchitecture.X86:
            return X86Info(c_cpu_id)
        if cpu_info.architecture == enums.CPUArchitecture.ARM:
            return ARMInfo(c_cpu_id)
        return cpu_info

    @classmethod
    def from_current_cpu(cls):
        """Creates a :class:`CPUInfo` instance by identifying the current CPU."""
        c_cpu_id = ffi.new("struct cpu_id_t *")
        if lib.cpu_identify(ffi.NULL, c_cpu_id) != 0:
            raise CLibraryError
        return cls.from_c(c_cpu_id)

    @classmethod
    def from_purpose(cls, purpose: enums.CPUPurpose):
        """Creates a :class:`CPUInfo` instance by identifying a CPU with the requested purpose."""
        c_cpu_id = ffi.new("struct cpu_id_t *")
        if lib.cpu_request_core_type(purpose, ffi.NULL, c_cpu_id) != 0:
            raise CLibraryError
        return CPUInfo.from_c(c_cpu_id)

    @classmethod
    def from_raw(cls, raw_data: CPURawData):
        """Creates a :class:`CPUInfo` instance from an instance of :class:`CPURawData`"""
        c_cpu_id = ffi.new("struct cpu_id_t *")
        if lib.cpu_identify(raw_data.c_cpu_raw_data, c_cpu_id) != 0:
            raise CLibraryError
        return CPUInfo.from_c(c_cpu_id)

    @classmethod
    def from_raw_array_and_purpose(
        cls, raw_data_array: CPURawDataArray, purpose: enums.CPUPurpose
    ):
        """Identifies a CPU with a given purpose, returning a :class:`CPUInfo` instance."""
        c_cpu_id = ffi.new("struct cpu_id_t *")
        if (
            lib.cpu_request_core_type(
                purpose, raw_data_array.c_cpu_raw_data_array, c_cpu_id
            )
            != 0
        ):
            raise CLibraryError
        return CPUInfo.from_c(c_cpu_id)

    @property
    def architecture(self) -> enums.CPUArchitecture:
        """The CPU Architecture."""
        try:
            return enums.CPUArchitecture(self._c_cpu_id.architecture)
        except ValueError:
            return enums.CPUArchitecture.UNKNOWN

    @property
    def feature_level(self) -> enums.CPUFeatureLevel:
        """The CPU feature level."""
        try:
            return enums.CPUFeatureLevel(self._c_cpu_id.feature_level)
        except ValueError:
            return enums.CPUFeatureLevel.UNKNOWN

    @property
    def vendor_str(self) -> str:
        """The CPU vendor string, e.g., :const:`'GenuineIntel'`."""
        return c_string_to_str(self._c_cpu_id.vendor_str)

    @property
    def brand_str(self) -> str:
        """The CPU brand string, e.g., :const:`'Intel(R) Xeon(TM) CPU 2.40GHz'`."""
        return c_string_to_str(self._c_cpu_id.brand_str)

    @property
    def vendor(self) -> enums.CPUVendor:
        """The CPU vendor."""
        try:
            return enums.CPUVendor(self._c_cpu_id.vendor)
        except ValueError:
            return enums.CPUVendor.UNKNOWN

    @property
    def features(self) -> set[enums.CPUFeature]:
        """The set of supported CPU features."""
        if self._features is None:
            self._features = {
                flag
                for flag in enums.CPUFeature
                if flag >= 0 and self._c_cpu_id.flags[flag]
            }
        return self._features

    @property
    def num_cores(self) -> int:
        """The number of CPU cores."""
        return self._c_cpu_id.num_cores

    @property
    def num_logical_cpus(self) -> int:
        """The number of logical processors."""
        return self._c_cpu_id.num_logical_cpus

    @property
    def total_logical_cpus(self) -> int:
        """The total number of logical processors."""
        return self._c_cpu_id.total_logical_cpus

    @property
    def l1_data_cache(self) -> Optional[int]:
        """L1 data cache size in KB. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l1_data_cache)

    @property
    def l1_instruction_cache(self) -> Optional[int]:
        """L1 instruction cache size in KB. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l1_instruction_cache)

    @property
    def l2_cache(self) -> Optional[int]:
        """L2 cache size in KB. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l2_cache)

    @property
    def l3_cache(self) -> Optional[int]:
        """L3 cache size in KB. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l3_cache)

    @property
    def l4_cache(self) -> Optional[int]:
        """L4 cache size in KB. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l4_cache)

    @property
    def l1_data_assoc(self) -> Optional[int]:
        """L1 data cache associativity. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l1_data_assoc)

    @property
    def l1_instruction_assoc(self) -> Optional[int]:
        """L1 instruction cache associativity. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l1_instruction_assoc)

    @property
    def l2_assoc(self) -> Optional[int]:
        """L2 cache associativity. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l2_assoc)

    @property
    def l3_assoc(self) -> Optional[int]:
        """L3 cache associativity. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l3_assoc)

    @property
    def l4_assoc(self) -> Optional[int]:
        """L4 cache associativity. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l4_assoc)

    @property
    def l1_data_cacheline(self) -> Optional[int]:
        """L1 data cache line size. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l1_data_cacheline)

    @property
    def l1_instruction_cacheline(self) -> Optional[int]:
        """L1 instruction cache line size. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l1_instruction_cacheline)

    @property
    def l2_cacheline(self) -> Optional[int]:
        """L2 cache line size. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l2_cacheline)

    @property
    def l3_cacheline(self) -> Optional[int]:
        """L3 cache line size. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l3_cacheline)

    @property
    def l4_cacheline(self) -> Optional[int]:
        """L4 cache line size. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l4_cacheline)

    @property
    def l1_data_instances(self) -> Optional[int]:
        """Number of L1 data cache instances. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l1_data_instances)

    @property
    def l1_instruction_instances(self) -> Optional[int]:
        """Number of L1 instruction cache instances. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l1_instruction_instances)

    @property
    def l2_instances(self) -> Optional[int]:
        """Number of L2 cache instances. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l2_instances)

    @property
    def l3_instances(self) -> Optional[int]:
        """Number of L3 cache instances. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l3_instances)

    @property
    def l4_instances(self) -> Optional[int]:
        """Number of L4 cache instances. :const:`None` if not determined."""
        return optional_int(self._c_cpu_id.l4_instances)

    @property
    def cpu_codename(self) -> str:
        """A human-friendly CPU codename."""
        return c_string_to_str(self._c_cpu_id.cpu_codename)

    @property
    def detection_hints(self) -> set[enums.CPUHint]:
        """A set of miscellaneous detection hints."""
        if self._detection_hints is None:
            self._detection_hints = {
                hint
                for hint in enums.CPUHint
                if hint >= 0 and self._c_cpu_id.detection_hints[hint]
            }
        return self._detection_hints

    @property
    def affinity_mask(self) -> bytes:
        """A bit mask of the affinity IDs that this processor type is occupying."""
        bit_mask = getattr(self._c_cpu_id.affinity_mask, "__bits")
        return b"".join(reversed([byte.to_bytes(1) for byte in bit_mask])).lstrip(
            b"\x00"
        )

    @property
    def purpose(self) -> enums.CPUPurpose:
        """The purpose of the CPU type, relevant for hybrid CPUs."""
        return enums.CPUPurpose(self._c_cpu_id.purpose)

    @property
    def hypervisor(self) -> Optional[enums.HypervisorVendor]:
        """The hypervisor vendor or :const:`None` if not detected."""
        hypervisor = lib.cpuid_get_hypervisor(ffi.NULL, self._c_cpu_id)
        return enums.HypervisorVendor(hypervisor) if hypervisor != 0 else None


class X86Info(CPUInfo):
    """
    The :class:`CPUInfo` child class for x86 CPUs.
    """

    def __init__(self, c_cpu_id):
        super().__init__(c_cpu_id)
        self._c_cpu_id = c_cpu_id

    @property
    def family(self) -> int:
        """The CPU family number."""
        return self._c_cpu_id.family

    @property
    def model(self) -> int:
        """The CPU model number."""
        return self._c_cpu_id.model

    @property
    def stepping(self) -> int:
        """The CPU stepping."""
        return self._c_cpu_id.stepping

    @property
    def ext_family(self) -> int:
        """The CPU display family number."""
        return self._c_cpu_id.ext_family

    @property
    def ext_model(self) -> int:
        """The CPU display model number."""
        return self._c_cpu_id.ext_model

    @property
    def sse_size(self) -> Optional[int]:
        """SSE execution unit size (64 or 128), :const:`None` if not available."""
        return optional_int(self._c_cpu_id.sse_size)

    @property
    def sgx(self) -> Optional[SGX]:
        """SGX-related features if present, otherwise :const:`None`."""
        return SGX(self._c_cpu_id.sgx) if self._c_cpu_id.sgx.present == 1 else None


class ARMInfo(CPUInfo):
    """
    The :class:`CPUInfo` child class for ARM CPUs.
    """

    def __init__(self, c_cpu_id):
        super().__init__(c_cpu_id)
        self._c_cpu_id = c_cpu_id

    @property
    def implementer(self) -> int:
        """The CPU implementer code."""
        return self._c_cpu_id.implementer

    @property
    def variant(self) -> int:
        """The CPU variant number."""
        return self._c_cpu_id.variant

    @property
    def part_num(self) -> int:
        """The CPU primary part number."""
        return self._c_cpu_id.part_num

    @property
    def revision(self) -> int:
        """The CPU revision number."""
        return self._c_cpu_id.revision


class SystemInfo:
    """
    Class for holding structured :class:`CPUInfo` information about multiple CPUs.
    Instances of this class can be indexed like lists, each item (of type :class:`CPUInfo`)
    corresponds to a different :class:`CPUPurpose`.
    """

    def __init__(self, c_system_id, cpu_info_list: list[CPUInfo]):
        self._c_system_id = c_system_id
        self._cpu_info_list = cpu_info_list

    def __del__(self):
        lib.cpuid_free_system_id(self._c_system_id)

    def __getitem__(self, index: int) -> CPUInfo:
        return self._cpu_info_list[index]

    def __len__(self) -> int:
        return len(self._cpu_info_list)

    @classmethod
    def from_c(cls, c_system_id):
        """Create a :class:`SystemInfo` instance from the corresponding C structure."""
        cpu_id_list = []
        for c_cpu_id in c_system_id.cpu_types[0 : c_system_id.num_cpu_types]:
            new_c_cpu_id = ffi.new("struct cpu_id_t *")
            ffi.memmove(
                new_c_cpu_id, ffi.addressof(c_cpu_id), ffi.sizeof("struct cpu_id_t")
            )
            cpu_id_list.append(CPUInfo.from_c(new_c_cpu_id))
        return cls(c_system_id, cpu_id_list)

    @classmethod
    def from_all_cpus(cls):
        """Create a :class:`SystemInfo` instance by indentifying all CPUs."""
        c_system_id = ffi.new("struct system_id_t *")
        if lib.cpu_identify_all(ffi.NULL, c_system_id) != 0:
            raise CLibraryError
        return cls.from_c(c_system_id)

    @classmethod
    def from_raw_array(cls, raw_data_array: CPURawDataArray):
        """Create a :class:`SystemInfo` instance from an instance of :class:`CPURawDataArray`"""
        c_system_id = ffi.new("struct system_id_t *")
        if lib.cpu_identify_all(raw_data_array.c_cpu_raw_data_array, c_system_id) != 0:
            raise CLibraryError
        return SystemInfo.from_c(c_system_id)

    @property
    def l1_data_total_instances(self) -> Optional[int]:
        """The number of total L1 data cache instances. :const:`None` if not determined."""
        return optional_int(self._c_system_id.l1_data_total_instances)

    @property
    def l1_instruction_total_instances(self) -> Optional[int]:
        """The number of total L1 instruction cache instances. :const:`None` if not determined."""
        return optional_int(self._c_system_id.l1_instruction_total_instances)

    @property
    def l2_total_instances(self) -> Optional[int]:
        """The number of total L2 cache instances. :const:`None` if not undetermined"""
        return optional_int(self._c_system_id.l2_total_instances)

    @property
    def l3_total_instances(self) -> Optional[int]:
        """The number of total L3 cache instances. :const:`None` if not undetermined"""
        return optional_int(self._c_system_id.l3_total_instances)

    @property
    def l4_total_instances(self) -> Optional[int]:
        """The number of total L4 cache instances. :const:`None` if not undetermined"""
        return optional_int(self._c_system_id.l4_total_instances)



"""
Module for CPU clock/frequency calculation.
"""

from enum import Enum, auto
from libcpuid.errors import LibcpuidError
from libcpuid._libcpuid_cffi import (  # pylint: disable=no-name-in-module, import-error
    lib,
    ffi,
)


def exec_rdtsc() -> int:
    """
    Exectutes the `rdtsc` (read timestamp counter) instruction
    and returns the value of the counter.
    """
    result = ffi.new("uint64_t *")
    lib.cpu_rdtsc(result)
    return result[0]


class TSC:
    """
    Class for calculating clock frequency using differences
    in time and in the timestamp counter.
    """

    class State(Enum):
        """State of the measurement."""

        NOT_STARTED = auto()
        STARTED = auto()
        STOPPED = auto()

    def __init__(self):
        self._c_cpu_mark = ffi.new("struct cpu_mark_t *")
        self._state = self.State.NOT_STARTED

    @property
    def state(self) -> State:
        """Returns the current state of the measurement."""
        return self._state

    def start(self):
        """Starts the measurement."""
        lib.cpu_tsc_mark(self._c_cpu_mark)
        self._state = self.State.STARTED

    def stop(self):
        """Stops the measurement."""
        if self._state == self.State.NOT_STARTED:
            raise LibcpuidError("The measurement has not been started.")
        if self._state == self.State.STOPPED:
            raise LibcpuidError("The measurement has already been stopped.")
        lib.cpu_tsc_unmark(self._c_cpu_mark)
        self._state = self.State.STOPPED

    def frequency(self) -> int:
        """
        Calculates the CPU clock frequency from the time and TSC
        difference between the calls to :meth:`start` and :meth:`stop`.
        """
        if self._state == self.State.NOT_STARTED:
            raise LibcpuidError("The measurementhas not been started.")
        if self._state == self.State.STARTED:
            raise LibcpuidError("The measurement has not been stopped.")
        result = lib.cpu_clock_by_mark(self._c_cpu_mark)
        if result < 0:
            raise LibcpuidError("Could not compute clock frequency.")
        return result


def frequency_measure(millis: int = 100, quad_check: bool = False) -> int:
    """
    Measures the CPU clock frequency by busy-waiting for a specified
    number of milliseconds and calculating the difference between TSC
    values. With the :const:`quad_check` option on, the function will
    run 4 consecutive measurements and return the average of the two
    most consistent results.
    """
    result = lib.cpu_clock_measure(millis, int(quad_check))
    if result < 0:
        raise LibcpuidError("Could not measure clock frequency.")
    return result


def frequency_by_os() -> int:
    """Returns the CPU clock frequency as reported by the OS."""
    result = lib.cpu_clock_by_os()
    if result < 0:
        raise LibcpuidError("Clock frequency unavailable.")
    return result


def frequency_by_ic(millis: int = 50, runs: int = 4) -> int:
    """
    Measures the CPU clock frequency using instruction counting for
    a specified number of milliseconds and runs.
    """
    result = lib.cpu_clock_by_ic(millis, runs)
    if result < 0:
        raise LibcpuidError("Could not measure clock frequency.")
    return result


def frequency_by_tsc() -> int:
    """Computes the CPU clock **base** frequency based on information from `cpuid`."""
    result = lib.cpu_clock_by_tsc(ffi.NULL)
    if result < 0:
        raise LibcpuidError("Could not compute clock frequency.")
    return result


def frequency() -> int:
    """
    All in one method for getting CPU clock frequency. It tries to use the OS
    and falls back to other methods if the OS does not have the information.
    """
    result = lib.cpu_clock()
    if result < 0:
        raise LibcpuidError("Could not get clock frequency.")
    return result



"""
Module providing access to enums from the libcpuid library,
e.g., CPU architectures, vendors, or features.

For compatibility with the underlying C library, the values
of all enums are integers. However, calling the :func:`str`
or :func:`print` function on an enum member returns or prints
its string representation.
"""

from enum import IntEnum
from libcpuid._utils import c_string_to_str, get_enum_options
from libcpuid._libcpuid_cffi import (  # pylint: disable=no-name-in-module, import-error
    lib,
)


CPUArchitecture = IntEnum("CPUArchitecture", get_enum_options("ARCHITECTURE_"))
CPUArchitecture.__str__ = lambda self: c_string_to_str(lib.cpu_architecture_str(self))
CPUArchitecture.__doc__ = "CPU architectures."

CPUVendor = IntEnum("CPUVendor", get_enum_options("VENDOR_"))
CPUVendor.__str__ = lambda self: self.name
CPUVendor.__doc__ = "CPU vendors."

CPUPurpose = IntEnum("CPUPurpose", get_enum_options("PURPOSE_"))
CPUPurpose.__str__ = lambda self: c_string_to_str(lib.cpu_purpose_str(self))
CPUPurpose.__doc__ = "CPU purposes."

HypervisorVendor = IntEnum(
    "HypervisorVendor", get_enum_options("HYPERVISOR_", ignore=["NONE"])
)
HypervisorVendor.__str__ = lambda self: self.name
HypervisorVendor.__doc__ = (
    "Hypervisor vendors, as guessed from the CPU_FEATURE_HYPERVISOR flag"
)

CPUFeature = IntEnum("CPUFeature", get_enum_options("CPU_FEATURE_"))
CPUFeature.__str__ = lambda self: c_string_to_str(lib.cpu_feature_str(self))
CPUFeature.__doc__ = "CPU feature identifiers (flags)"

CPUFeatureLevel = IntEnum("CPUFeatureLevel", get_enum_options("FEATURE_LEVEL_"))
CPUFeatureLevel.__str__ = lambda self: c_string_to_str(lib.cpu_feature_level_str(self))
CPUFeatureLevel.__doc__ = """
    CPU feature levels, also known as microarchitecture levels (x86) or architecture versions (ARM).
"""

CPUHint = IntEnum("CPUHint", get_enum_options("CPU_HINT_"))
CPUHint.__str__ = lambda self: self.name
CPUHint.__doc__ = "CPU detection hint identifiers."

SGXFeature = IntEnum("SGXFeature", get_enum_options("INTEL_"))
SGXFeature.__str__ = lambda self: self.name
SGXFeature.__doc__ = "SGX feature flags."



#!/usr/bin/env python3

import argparse, sys, re


### Constants:
delimiter = "-" * 80

def readRawFile():
	rawdata = []
	for line in args.raw_file.readlines():
		lookfor = [
			"Logical CPU", "CPUID", "CPU#", # common
			"basic_cpuid", "ext_cpuid", "intel_fn4", "intel_fn11", "amd_fn8000001dh", # x86
			"arm_midr", "arm_mpidr", "arm_revidr", # ARM common
			"arm_id_afr", "arm_id_dfr", "arm_id_isar", "arm_id_mmfr", "arm_id_pfr", # ARM (AArch32)
			"arm_id_aa64afr", "arm_id_aa64dfr", "arm_id_aa64fpfr", "arm_id_aa64isar", "arm_id_aa64mmfr", "arm_id_aa64pfr", "arm_id_aa64smfr", "arm_id_aa64zfr" # ARM (AArch64)
		]
		ignore  = ["Manufacturer", "CPU Name", "Revision", "MSR Register"]
		good = False
		for match in lookfor:
			if line.find(match) != -1:
				good = True
				break
		for match in ignore:
			if line.find(match) != -1:
				good = False
				break
		if good:
			rawdata.append(line.strip())
	return rawdata

def readResultFile():
	repdata = []
	rexp = re.compile('(-?[0-9]+).*')
	for line in args.report_file.readlines():
		s = line.strip()
		if s.find(":") == -1:
			continue
		numeric = ["family", "model", "stepping", "ext_family", "ext_model",
				"implementer", "variant", "part_num", "revision",
				"num_cores", "num_logical",
				"L1 D cache",  "L1 I cache",  "L2 cache",   "L3 cache",   "L4 cache",
				"L1D assoc.",  "L1I assoc.",  "L2 assoc.",  "L3 assoc.",  "L4 assoc.",
				"L1D line sz", "L1I line sz", "L2 line sz", "L3 line sz", "L4 line sz",
				"L1D inst.",   "L1I inst.",   "L2 inst.",   "L3 inst.",   "L4 inst.",
				]
		field = s[:s.find(":")].strip()
		if field in numeric:
			value = s[s.find(":")+1:].strip()
			if not rexp.match(value):
				raise f"Bad format of value: [{s}]"
			repdata.append(rexp.findall(value)[0])
		if "CPU Info for type" in field:
			repdata.append(delimiter)
		if field == "arch":
			value = s[s.find(":") + 2:]
			repdata.append(value)
		if field == "feat_level":
			value = s[s.find(":") + 2:]
			repdata.append(value)
		if field == "purpose":
			value = s[s.find(":") + 2:]
			repdata.append(value)
		if field == "code name":
			value = s[s.find("`") + 1: s.find("'")]
			repdata.append(value)
		if field == "technology":
			value = s[s.find("`") + 1: s.find("'")]
			repdata.append(value)
		if field == "features":
			value = s[s.find(":") + 2:]
			repdata.append(value)
		if field == "SSE units":
			value = s[s.find(":") + 2:]
			# the value here is something like "XX bits (authoritative)". We remove the "bits" part:
			i = value.find("bits")
			if i != -1:
				value = value[:i] + value[i + 5:]
			repdata.append(value)
	return repdata

# Parse arguments
parser = argparse.ArgumentParser(description="Create a new test file by using cpuid_tool raw and report files.")
parser.add_argument("raw_file",
	nargs='?',
	type=argparse.FileType('r'),
	default="raw.txt",
	help="an existing raw data file")
parser.add_argument("report_file",
	nargs='?',
	type=argparse.FileType('r'),
	default="report.txt",
	help="an existing report file")
parser.add_argument("test_file",
	nargs='?',
	type=argparse.FileType('w'),
	default=sys.stdout,
	help="test file to create (default is standard output)")
args = parser.parse_args()

# Create test file
lines = readRawFile() + readResultFile()
args.test_file.writelines([s + "\n" for s in lines])



"""
Module dealing with SGX-related CPU information.
"""

from libcpuid import enums
from libcpuid._libcpuid_cffi import (  # pylint: disable=no-name-in-module, import-error
    lib,
    ffi,
)


class EPC:
    """
    Class holding information about a single EPC (Enclave Page Cache) area.
    """

    def __init__(self, c_epc):
        self._c_epc = c_epc

    @property
    def start_addr(self) -> int:
        """Start address."""
        return self._c_epc.start_addr

    @property
    def length(self) -> int:
        """Length of the area."""
        return self._c_epc.length


class SGX:
    """
    Class holding the SGX-related information about a CPU.
    """

    def __init__(self, c_sgx):
        self._c_sgx = c_sgx
        self._features = None

    @property
    def max_enclave_32bit(self) -> int:
        """The maximum enclave size in 32-bit mode."""
        return self._c_sgx.max_enclave_32bit

    @property
    def max_enclave_64bit(self) -> int:
        """The maximum enclave size in 64-bit mode."""
        return self._c_sgx.max_enclave_64bit

    @property
    def features(self) -> set[enums.SGXFeature]:
        """Set of CPU SGX features."""
        if self._features is None:
            self._features = {
                flag
                for flag in enums.SGXFeature
                if flag >= 0 and self._c_sgx.flags[flag]
            }
        return self._features

    @property
    def num_epc_sections(self) -> int:
        """The number of Enclave Page Cache (EPC) sections."""
        return self._c_sgx.num_epc_sections

    @property
    def misc_select(self) -> bytes:
        """
        Bit vector of the supported extended features that can be written
        to the MISC region of the SSA (Save State Area)
        """
        return self._c_sgx.misc_select.to_bytes(4)

    @property
    def secs_attributes(self) -> bytes:
        """
        Bit vector of the attributes that can be set to SECS.ATTRIBUTES
        via ECREATE.
        """
        return self._c_sgx.secs_attributes.to_bytes(8)

    @property
    def secs_xfrm(self) -> bytes:
        """
        Bit vector of the attributes that can be set in the XSAVE feature
        request mask.
        """
        return self._c_sgx.secs_xfrm.to_bytes(8)

    def get_epc(self, index: int) -> EPC:
        """Fetches information about a single EPC area given by `index`."""
        if not 0 <= index < self.num_epc_sections:
            raise IndexError(
                f"Invalid index {index}, the CPU has {self.num_epc_sections} EPC sections."
            )
        return EPC(lib.cpuid_get_epc(index, ffi.NULL))



"""
Module containing custom exceptions of the library.
"""

from libcpuid._utils import c_string_to_str
from libcpuid._libcpuid_cffi import (  # pylint: disable=no-name-in-module, import-error
    lib,
)


class LibcpuidError(Exception):
    """Main exception of the library."""


class CLibraryError(LibcpuidError):
    """
    Raised when an error code is returned by some function
    in the underlying libcpuid C library.
    """

    def __init__(self):
        super().__init__(c_string_to_str(lib.cpuid_error()))



"""
Internal module containing utility functions.
"""

from typing import Optional
from libcpuid._libcpuid_cffi import (  # pylint: disable=no-name-in-module, import-error
    ffi,
    lib,
)


def c_string_to_str(c_string) -> str:
    """Converts an FFI C string to a Python string."""
    return ffi.string(c_string).decode()


def optional_int(value: int) -> Optional[int]:
    """Returns the given integer if it is not -1, otherwise None."""
    return value if value != -1 else None


def get_enum_options(  # pylint: disable=dangerous-default-value
    prefix, ignore=[]
) -> dict[str, int]:
    """Returns a dictionary of C enum option names and their values based on name prefix."""
    enum_options = [x for x in dir(lib) if x.startswith(prefix)]
    return {
        x[len(prefix) :]: getattr(lib, x)
        for x in enum_options
        if x[len(prefix) :] not in ignore
    }



"""
Module providing access to information from model-specific registers.
"""

from typing import Optional
from libcpuid.errors import CLibraryError
from libcpuid._utils import get_enum_options
from libcpuid._libcpuid_cffi import (  # pylint: disable=no-name-in-module, import-error
    lib,
    ffi,
)


class MSR:
    """
    Class for reading information from model-specific registers (MSR).

    Since `rdmsr` is a privileged instruction, you need ring 0 access to
    instantiate this class.

    To see the documentation of properties, look for :const:`cpu_msrinfo_request_t`
    in :const:`libcpuid.h`, the main header file of the underlying C library.
    """

    def __init__(self, core: int = 0):
        self._handle = lib.cpu_msr_driver_open_core(core)
        if self._handle == ffi.NULL:
            raise CLibraryError

    def __del__(self):
        if lib.cpu_msr_driver_close(self._handle) != 0:
            raise CLibraryError

    def serialize_raw_data(self, filename: str):
        """Writes raw MSR data to a file."""
        if lib.msr_serialize_raw_data(self._handle, filename.encode()) != 0:
            raise CLibraryError

    def read_register(self, msr_index: int) -> bytes:
        """Returns the content of a model-specific register with the given index."""
        result = ffi.new("uint64_t *")
        if lib.cpu_rdmsr(self._handle, msr_index, result) != 0:
            raise CLibraryError
        return result[0].to_bytes(8)


def _get_msrinfo_request(enum_value):
    def _msrinfo_request(self) -> Optional[int]:
        """May return :const:`None` if not available."""
        result = lib.cpu_msrinfo(
            self._handle, enum_value  # pylint: disable=protected-access
        )
        return None if result == 0x3FFFFFFF else result

    return _msrinfo_request


# This adds properties to the MSR class, copying the possible
# inputs to the cpu_msrinfo function of the C library, such as
# mperf, aperf, min_multiplier, temperature, or voltage.
# See cpu_msr_info_request_t in libcpuid.h for details.
for name, value in get_enum_options("INFO_", ignore=["BCLK"]).items():
    setattr(MSR, name.lower(), property(_get_msrinfo_request(value)))



#!/usr/bin/env python3

import argparse, os, sys, re, glob
from pathlib import Path


### Constants:
git_root_dir = os.popen("git rev-parse --show-toplevel").read().splitlines()

err = 0
def getEnumElements(enumName):
	f = open(f"{args.root_dir}/libcpuid.h", "r")
	l = []
	on = False
	rexp = re.compile(r'^\s*([A-Z0-9_]+)(\s*=\s*[A-Z0-9_]+)?\s*,.*$')
	for line in f:
		line = line.strip()
		if line.startswith("typedef enum {"):
			l = []
			on = True
		if on and rexp.match(line):
			l.append(rexp.findall(line)[0][0])
		if on and line.startswith("} "):
			on = False
			if line.find(enumName) != -1:
				return l
	f.close()
	return []

def getConstant(constantName):
	f = open(f"{args.root_dir}/libcpuid_constants.h", "r")
	value = 0
	for line in f:
		items = line.strip().split()
		if len(items) >= 3 and items[0] == "#define" and items[1] == constantName:
			value = int(items[2])
	f.close()
	return value

def checkEnumSize(enumName, constantName):
	print(f"Checking enum `{enumName}':", end=' ')
	count = len(getEnumElements(enumName)) - 1
	themax = getConstant(constantName)
	print(f"{count} elements; max size ({constantName}={themax})...", end=' ')
	if count > themax:
		err += 1
		print("FAILED")
		global firstError
		firstError = False
	else:
		print("OK")

def check_type_directory(directory):
	if not Path(directory).is_dir():
		raise argparse.ArgumentTypeError(f"{directory} is not a directory")
	return directory

# Parse arguments
parser = argparse.ArgumentParser(description="Check library consistency.")
parser.add_argument("root_dir",
	nargs='?',
	type=check_type_directory,
	default=f"{git_root_dir[-1]}/libcpuid" if len(git_root_dir) > 0 else "",
	help="path to the libcpuid sub-directory")
args = parser.parse_args()

checkEnumSize("cpu_feature_t", "CPU_FLAGS_MAX")
checkEnumSize("cpu_hint_t", "CPU_HINTS_MAX")
checkEnumSize("cpu_sgx_feature_t", "SGX_FLAGS_MAX")

rexp = re.compile('.*{ CPU_FEATURE_([^,]+), "([^"]+)".*}.*')
print("Finding features:")
for fn in glob.glob(f"{args.root_dir}/*.c"):
	f = open(fn, "rt")
	line = 1
	nfeat = 0
	for s in f.readlines():
		if rexp.match(s) and "FEATURE_LEVEL_" not in s:
			nfeat += 1
			res = rexp.findall(s)
			assert len(res) == 1, "Too many matches"
			if res[0][0].lower() != res[0][1]:
				err += 1
				print(f"..Mismatch - {os.path.basename(fn)}:{line} - `{res[0][0]}' vs `{res[0][1]}'")
		line += 1
	if nfeat:
		print(f"  {os.path.basename(fn)}: {nfeat} features described")
	f.close()

# Check whether all features are converted by cpu_feature_str():

allf = []
f = open(f"{args.root_dir}/libcpuid.h", "rt")
rexp = re.compile('\t(CPU_FEATURE_[^, ]+).*')
for s in f.readlines():
	if rexp.match(s):
		allf.append(rexp.findall(s)[0])

f.close()

impf = []
rexp = re.compile('\t+{ (CPU_FEATURE_[^,]+).*')
f = open(f"{args.root_dir}/cpuid_main.c", "rt")
for s in f.readlines():
	if rexp.match(s):
		entry = rexp.findall(s)[0]
		if entry in impf:
			err += 1
			print(f"cpu_feature_str(): duplicate entry: {entry}")
		impf.append(entry)
f.close()

print(f"Found {len(allf)} total features and {len(impf)} named features")

for feature in allf:
	if not feature in impf:
		err += 1
		print(f"cpu_feature_str(): don't have entry for {feature}")

# Check whether all features have detection code:
print("Checking whether all features have detection code...", end=' ')
firstError = True

files_code = {}

rexp1 = re.compile(r'.*\[(CPU_FEATURE_[^ \]]+)\]\s*=\s*{.*') # e.g. "[CPU_FEATURE_SWAP] = {"
rexp2 = re.compile(r'.*(CPU_FEATURE_[^ ,]+),\s*FEATURE_LEVEL_ARM_.*') # e.g. "set_feature_status(data, ext_status, (mte_frac == 0b0000), CPU_FEATURE_MTE_ASYNC,           FEATURE_LEVEL_ARM_V8_5_A, -1);"
rexp3 = re.compile(r'.*(CPU_FEATURE_[^ }]+).*') # e.g. "{ 28, CPU_FEATURE_HT },"

for fn in glob.glob(f"{args.root_dir}/*.c"):
	f = open(fn, "rt")
	files_code[fn] = []
	for s in f.readlines():
		if rexp1.match(s):
			entry = rexp1.findall(s)[0]
			files_code[fn].append(entry)
		elif rexp2.match(s):
			entry = rexp2.findall(s)[0]
			files_code[fn].append(entry)
		elif rexp3.match(s):
			entry = rexp3.findall(s)[0]
			files_code[fn].append(entry)

	f.close()

features_whitelist = [
	"CPU_FEATURE_SSE5", # deprecated, never defined
	"CPU_FEATURE_AES", # defined twice (x86 + ARM)
]
for feature in allf:
	matching_files = []
	for fn in files_code:
		if feature in files_code[fn]:
			matching_files.append(fn)
	if len(matching_files) == 0 and feature not in features_whitelist:
		if firstError:
			print("FAILED:")
			firstError = False
		err += 1
		print(f"..No detection code for {feature}")
	if len(matching_files) > 1 and feature not in features_whitelist:
		if firstError:
			print("FAILED:")
			firstError = False
		err += 1
		print(f"..Conflicting detection code for {feature} in files {' and '.join(matching_files)}")

if firstError:
	print("All OK.")
print("")

print("Checking processor definitions:")
cache_exp = re.compile(r".*([\(/ ][0-9]+K).*")
# Check whether CPU codenames for consistency:
#   - Codenames should not exceed 31 characters
#   - Check for common typos
definitions = 0
match_entry_fields = 12 # this number needs to change if the definition of match_entry_t ever changes
codename_str_max = 64-1 # this number needs to change if the value of CODENAME_STR_MAX ever changes
technology_str_max = 16-1 # this number needs to change if the value of TECHNOLOGY_STR_MAX ever changes
common_cache_sizes = ["8", "16", "32", "64", "128", "256", "512", "1024", "2048", "3072", "4096", "6144", "8192", "12288", "16384"]
for fn in glob.glob(f"{args.root_dir}/*.c"):
	bfn = os.path.basename(fn)
	nline = 0
	f = open(fn, "rt")
	has_matchtable = False
	cdefs = 0
	allok = True
	for line in f.readlines():
		nline += 1
		if line.find("struct match_entry_t") != -1:
			has_matchtable = True
		if not has_matchtable:
			continue
		i = line.find("{")
		j = line.rfind("}")
		if i == -1 or j == -1 or i > j:
			continue
		inner = line[i+1:j]
		parts = inner.split(",")
		if len(parts) == match_entry_fields:
			cdefs += 1
			# Check CPU codename string
			codename = parts[10].strip()
			if codename[0] != '"' or codename[-1] != '"':
				err += 1
				print(f"..Warning, {bfn}:{nline} - cannot correctly handle the cpu codename ({codename})")
				allok = False
				continue
			codename = codename[1:-1]
			if len(codename) > codename_str_max:
				err += 1
				print(f"..{bfn}:{nline} - codename ({codename}) is longer than {codename_str_max} characters!")
				allok = False
			if cache_exp.match(codename):
				cache_size = cache_exp.findall(codename)[0][1:-1]
				if not cache_size in common_cache_sizes:
					err += 1
					print(f"..Warning, {bfn}:{nline} - suspicious cache size in codename [{codename}] ({cache_size})")
					allok = False
			# Check CPU technology string
			technology = parts[11].strip()
			if (technology[0] != '"' or technology[-1] != '"') and technology != "UNKN_STR":
				err += 1
				print(f"..Warning, {bfn}:{nline} - cannot correctly handle the cpu technology ({technology})")
				allok = False
				continue
			technology = technology[1:-1]
			if len(technology) > technology_str_max:
				err += 1
				print(f"..{bfn}:{nline} - technology ({technology}) is longer than {technology_str_max} characters!")
				allok = False
	if cdefs:
		definitions += 1
		print(f"  {bfn}: {cdefs} processor definitions,", end=' ')
		if allok:
			print("all OK")
		else:
			print("some errors/warnings")
	f.close()
if definitions == 0:
	err += 1
	print("..Warning, no processor definitions found")

sys.exit(err)



"""Sanity tests for the libcpuid package."""

import os
import tempfile
import libcpuid
from libcpuid.info import CPUInfo, SystemInfo
from libcpuid.raw import CPURawData, CPURawDataArray
from libcpuid.errors import CLibraryError


def test_cpu_name_in_vendor_list():
    """
    Checks that the current CPU codename appears
    in the list of all CPUs of its vendor.
    """
    try:
        info = CPUInfo.from_current_cpu()
        cpulist = libcpuid.get_cpu_list(info.vendor)
        assert info.cpu_codename in cpulist
    except CLibraryError:
        pass


def test_serialization():
    """
    Checks that CPU data serialization and
    deserialization does not raise any errors.
    """
    with tempfile.TemporaryDirectory() as tmpdirname:
        info_file = os.path.join(tmpdirname, "info.txt")
        raw = CPURawData.from_current_cpu()
        raw.serialize(info_file)
        info = CPUInfo.from_raw(raw)
        info_from_file = CPUInfo.from_raw(CPURawData.from_file(info_file))
        assert info.features == info_from_file.features


def test_array_serialization():
    """
    Checks that CPU array data serialization and
    deserialization does not raise any errors.
    """
    with tempfile.TemporaryDirectory() as tmpdirname:
        info_file = os.path.join(tmpdirname, "info.txt")
        raw_array = CPURawDataArray.from_all_cpus()
        raw_array.serialize(info_file)
        info = SystemInfo.from_raw_array(raw_array)
        info_from_file = SystemInfo.from_raw_array(CPURawDataArray.from_file(info_file))
        assert len(info) == len(info_from_file)
        assert info[0].features == info_from_file[0].features



"""
Module for compiling the C FFI.
"""

import os
import subprocess
import tempfile
import re
import argparse
from pathlib import Path
from cffi import FFI


class FFIBuildException(Exception):
    """Generic exception for errors occuring during the CFFI build."""


def preprocess_header(header_path):
    """
    Preprocesses the header file (python-cffi only accepts preprocessed C definitions)
    at the given path and returns it as a string.
    """
    try:
        return subprocess.check_output(
            ["gcc", "-U __GNUC__", "-E", header_path]
        ).decode()
    except subprocess.CalledProcessError as e:
        if e.returncode == 127:
            raise FFIBuildException(
                "The gcc compiler is necessary to build python-libcpuid."
            ) from e
        raise FFIBuildException(
            f"Error preprocessing the libcpuid header file: {e.stderr}"
        ) from e


def _get_sizeof_eval_source(sizeof):
    return f"""
#include <libcpuid.h>
#include <stdio.h>

int main() {{
    printf("%ld", {sizeof});
    return 0;
}}
"""


def eval_sizeofs(header, cflags):
    """
    Evaluates each sizeof found in the given C header and replaces all
    occurences of the sizeof with its computed value.
    """
    sizeofs = set(re.findall(r"sizeof\([^\)]*\)", header))
    tmp_dir = tempfile.mkdtemp()
    c_program_path = Path(tmp_dir, "sizeof.c")
    executable_path = Path(tmp_dir, "sizeof")

    for sizeof in sizeofs:
        with open(c_program_path, "w", encoding="UTF-8") as c_program_file:
            c_program_file.write(_get_sizeof_eval_source(sizeof))
        subprocess.check_call(["gcc", c_program_path, *cflags, "-o", executable_path])
        size = subprocess.check_output([executable_path]).decode()
        header = header.replace(sizeof, size)

    os.remove(c_program_path)
    os.remove(executable_path)
    os.rmdir(tmp_dir)
    return header


LIBCPUID_DIR = str(Path(*(Path(os.path.abspath(__file__)).parts[:-4])))
LIBCPUID_INCLUDE_DIR = str(Path(LIBCPUID_DIR, "libcpuid"))
LIBCPUID_LIBRARY_DIR = str(Path(LIBCPUID_DIR, "libcpuid", ".libs"))
LIBCPUID_MAIN_HEADER_FILENAME = "libcpuid.h"
LIBCPUID_MAIN_HEADER_PATH = str(
    Path(LIBCPUID_INCLUDE_DIR, LIBCPUID_MAIN_HEADER_FILENAME)
)
LIBCPUID_LIBRARY_NAME = "cpuid"
PYTHON_SRC_DIR = str(Path(LIBCPUID_DIR, "python", "src"))

PREPROCESSED_HEADER = preprocess_header(LIBCPUID_MAIN_HEADER_PATH)
EVAL_SIZEOF_CFLAGS = [
    f"-I{LIBCPUID_INCLUDE_DIR}",
    f"-L{LIBCPUID_LIBRARY_DIR}",
    f"-l{LIBCPUID_LIBRARY_NAME}",
    f"-Wl,-rpath={LIBCPUID_LIBRARY_DIR}",
]

NO_SIZEOF_HEADER = eval_sizeofs(PREPROCESSED_HEADER, EVAL_SIZEOF_CFLAGS)

ffibuilder = FFI()
ffibuilder.cdef(NO_SIZEOF_HEADER)

set_source_kwargs = {
    "module_name": "libcpuid._libcpuid_cffi",
    "source": f"#include <{LIBCPUID_MAIN_HEADER_FILENAME}>",
    "libraries": [LIBCPUID_LIBRARY_NAME],
    "include_dirs": [LIBCPUID_INCLUDE_DIR],
    "library_dirs": [LIBCPUID_LIBRARY_DIR],
}

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-r", "--runtime-link", action="store_true")
    args = parser.parse_args()
    if args.runtime_link:
        set_source_kwargs["extra_link_args"] = [f"-Wl,-rpath={LIBCPUID_LIBRARY_DIR}"]

ffibuilder.set_source(**set_source_kwargs)

if __name__ == "__main__":
    ffibuilder.compile(PYTHON_SRC_DIR)



#!/usr/bin/env python3


import argparse, textwrap, os, sys, re, random, lzma
from pathlib import Path, PurePath


### Constants:
os.environ["LIBCPUID_NO_WARN"] = "1"
delimiter = "-" * 80
fields_x86 = [ "architecture", "feature-level", "purpose", "family", "model", "stepping",
	   "extfamily", "extmodel", "cores", "logical",
	   "l1d-cache",     "l1i-cache",     "l2-cache",     "l3-cache",     "l4-cache",
	   "l1d-assoc",     "l1i-assoc",     "l2-assoc",     "l3-assoc",     "l4-assoc",
	   "l1d-cacheline", "l1i-cacheline", "l2-cacheline", "l3-cacheline", "l4-cacheline",
	   "l1d-instances", "l1i-instances", "l2-instances", "l3-instances", "l4-instances",
	   "sse-size", "codename", "technology", "flags" ]
fields_arm = [ "architecture", "feature-level", "purpose",
	   "implementer", "variant", "part-num", "revision",
	   "cores", "logical",
	   "codename", "technology", "flags" ]

# One would usually use os.tempnam, but libc gives off hell a lot of
# warnings when you attempt to use that :(
def make_tempname(prefix):
	chars = ""
	for i in range(26):
		chars += chr(97+i)
		chars += chr(65+i)
	for i in range(10):
		chars += chr(48+i)
	for i in range(6):
		prefix += random.choice(chars)
	return prefix

def fmt_error(err):
	pfix = f"  {err[0]}: "
	return "{} expected `{}'\n{} got      `{}'".format(pfix, err[1], ' '*len(pfix), err[2])

def fixFile(filename, input_lines, output_lines):
	pfilename = Path(filename)
	if pfilename.suffixes[-1] == ".xz":
		# XZ compressed file
		f = lzma.open(pfilename, "wt")
	else:
		# Plain text file
		f = open(pfilename, "wt")
	f.writelines([s + "\n" for s in input_lines])
	f.write(delimiter + "\n")
	f.writelines([s + "\n" for s in output_lines])
	f.close()

def do_test(inp, expected_out, binary, test_file_name, num_cpu_type):
	fninp = make_tempname("cpuidin")
	fnoutp = make_tempname("cpuidout")
	f = open(fninp, "wt")
	f.writelines([s + "\n" for s in inp])
	f.close()
	architecture = os.popen(f"{binary} --load={fninp} --architecture").read().splitlines()[-1]
	if architecture == "x86":
		fields = fields_x86
	elif architecture == "ARM":
		fields = fields_arm
	else:
		fields = []
	cmd = f"{binary} --load={fninp} --outfile={fnoutp} {' '.join(['--' + s for s in fields])}"
	os.system(cmd)
	os.unlink(fninp)
	real_out = []
	real_out_delim = [] # when fixing the file, delimiters are required
	try:
		f = open(fnoutp, "rt")
		for s in f.readlines():
			real_out_delim.append(s.strip())
			if delimiter not in s:
				real_out.append(s.strip())
		f.close()
		os.unlink(fnoutp)
	except IOError:
		return "Exception"
	if len(real_out) != len(expected_out) or len(real_out) != len(fields) * num_cpu_type:
		if args.fix:
			fixFile(test_file_name, inp, real_out_delim)
			return "Number of records, fixed."
		else:
			return "Unexpected number of records returned\n - expected length {}\n - real length {}\n - {} fields".format(len(expected_out), len(real_out), len(fields) * num_cpu_type)
	err_fields = []
	for i in range(len(real_out)):
		if real_out[i] != expected_out[i]:
			err_fields.append((fields[i % len(fields)], expected_out[i], real_out[i]))
	if not err_fields:
		return "OK"
	else:
		if args.fix:
			fixFile(test_file_name, inp, real_out_delim)
			return "Mismatch, fixed."
		else:
			return "Mismatch in fields:\n{}".format('\n'.join([fmt_error(err) for err in err_fields]))

def is_regular_file(filename):
	try:
		with open(filename, 'r') as fd:
			fd.read()
			return True
	except:
		return False

def check_type_binary_file(filename):
	if not Path(filename).is_file():
		raise argparse.ArgumentTypeError(f"{filename} is not a file")
	if is_regular_file(filename):
		raise argparse.ArgumentTypeError(f"{filename} is not a binary file")
	return filename

# Parse arguments
parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
	description=textwrap.dedent("""

Run test files.

If a test file is given, it is tested by itself.
If a directory is given, process all *.test files there, subdirectories included.

If the --fix option is given, the behaviour of the cpuid_tool binary is deemed correct and any failing tests are updated.
"""))
parser.add_argument("cpuid_tool",
	nargs='?',
	type=check_type_binary_file,
	default="./build/cpuid_tool/cpuid_tool",
	help="path to the cpuid_tool binary")
parser.add_argument("input_test_files",
	nargs='+',
	default=["./tests"],
	help="test file or directory containing test files")
parser.add_argument("--fix",
	action=argparse.BooleanOptionalAction,
	default=False,
	help="update failing tests (default is false)")
parser.add_argument("--show-test-fast-warning",
	dest="show_test_fast_warning",
	action=argparse.BooleanOptionalAction,
	default=False,
	help="show a warning on errors (default if false)")
args = parser.parse_args()

# Create test files list
filelist = []
for input_test_file in args.input_test_files:
	if Path(input_test_file).is_dir():
		# gather all *.test files from subdirs amd, intel and cie:
		for dirpath, dirnames, filenames in os.walk(input_test_file):
			filelist += [PurePath(dirpath).joinpath(fn) for fn in filenames if ".test" in Path(fn).suffixes]
	else:
		filelist.append(input_test_file)

# Run tests
errors = False
print("Testing...")
for test_file_name_raw in filelist:
	test_file_name = Path(test_file_name_raw)
	num_cpu_type = 0
	current_input = []
	current_output = []
	build_output = False
	if test_file_name.suffixes[-1] == ".xz":
		# XZ compressed file
		f = lzma.open(test_file_name, "rt")
	else:
		# Plain text file
		f = open(test_file_name, "rt")
	# Read file line by line
	for line in f.readlines():
		line = line.strip()
		if line == delimiter:
			build_output = True
			num_cpu_type += 1
		else:
			if build_output:
				current_output.append(line)
			else:
				current_input.append(line)
	f.close()
	#codename = current_output[len(current_output) - 2]
	result = do_test(current_input, current_output, args.cpuid_tool, test_file_name, num_cpu_type)
	print(f"Test [{test_file_name.name}]: {result}")
	if result != "OK":
		errors = True
	build_output = False

if errors:
	if args.show_test_fast_warning:
		print("""
You're running tests in fast mode; before taking any action on the errors
above, please confirm that the slow mode ('make test-old') also fails.""")
	sys.exit(1)
else:
	print("All successfull!")
	sys.exit(0)
